

;*****************************************************************************
;                     Multi - Master Code for 83C751/83C752
;                                   4/14/1992
;*****************************************************************************

; This code was written to accompany an application note. The I2C routines
; are intended to be demonstrative and transportable into different
; application scenarios, and were NOT optimized for speed and/or memory
; utilization.


$TITLE(83C751 Multi Master I2C Routines)
$DATE(4/14/1992)
$MOD751
$DEBUG


;*****************************************************************************
;       8XC751 MULTI MASTER I2C COMMUNICATIONS ROUTINES
;       Symbols and RAM definitions
;*****************************************************************************

; Symbols (masks) for I2CFG bits.

BTIR       EQU     10h                 ; TIRUN bit.
BMRQ       EQU     40h                 ; MASTRQ bit.


; Symbols (masks) for I2CON bits.

BCXA       EQU     80h                 ; CXA bit.
BIDLE      EQU     40h                 ; IDLE bit.
BCDR       EQU     20h                 ; CDR bit.
BCARL      EQU     10h                 ; CARL bit.
BCSTR      EQU     08h                 ; CSTR bit.
BCSTP      EQU     04h                 ; CSTP bit.
BXSTR      EQU     02h                 ; XSTR bit.
BXSTP      EQU     01h                 ; XSTP bit.

; Note:

; Specific bits of the I2CON register are set by writing into this register a
; combination of the masks defined above using the MOV command.
; The SETB command should not be used with I2CON, as it is implemented by
; reading the contents of the register, setting the appropriate bit and
; writing  it back into the register.  As the functionality of the Read and
; Write  portions of the I2CON register is different, useing SETB may cause
; unwanted results.


; Message transaction status indications in MSGSTAT:

SGO        EQU     10h         ; Started Slave message processing.
SRCVD      EQU     11h         ; as a slave, received a new message
SRLNG      EQU     12h         ; received as slave a message which is too
                               ; long for the buffer
STXED      EQU     13h         ; as slave, completed message transmission.
SRERR      EQU     14h         ; bus error detected when operating as a slave.

MGO        EQU     20h         ; Started Master message processing.
MRCVED     EQU     21h         ; As Master, received complete message from
                               ; slave.
MTXED      EQU     22h         ; As Master, completed successful message
                               ; transmission (slave acknowledged all data
                               ; bytes).
MTXNAK     EQU     23h         ; As Master, truncated message since slave did
                               ; not acknowledge a data byte.
MTXNOSLV   EQU     24h         ; AS Master, did not receive an acknowledgement
                               ; for the specified slave address.

TIMOUT     EQU     30h         ; TIMERI Timed out.
NOTSTR     EQU     32h         ; Master did not recognize Start.

; RAM locations used by I2C interrupt service routines.

MASCMD     DATA    20h
SUBADD     BIT     MASCMD.0
RPSTRT     BIT     MASCMD.1
SETMRQ     BIT     MASCMD.2

           DSEG    AT 24h

MSGSTAT:   DS      1           ; I2C communications status.
MYADDR:    DS      1           ; Address of this I2C node.
DESTADRW:  DS      1           ; Destination address + R/W  (for Master).
DESSUBAD:  DS      1           ; Destination subaddress.
MASTCNT:   DS      1           ; Number of data bytes in message (Master,
                               ; send or receive).

TITOCNT:   DS      1           ; Timer I bus watchdog timeouts counter.
StackSave: DS      1           ; SP save location (used when returning from
                               ; bus recovery routine).

MasBuf:    DS      4           ; Master receive/transmit buffer, 8 bytes.
SRcvBuf:   DS      4           ; Slave receive buffer, 8 bytes.
STxBuf:    DS      4           ; Slave transmit buffer, 8 bytes.

RBufLen    EQU     4h          ; The length of SRcvBuf


;*****************************************************************************
;          APPLICATION output pins and RAM definitions
;*****************************************************************************

; Outputs used by the application:

TogLED     BIT     P1.0       ; Toggling output pin, to confirm
                              ; that the ping-pong game proceeds fine.
ErrLED     BIT     P1.1       ; Error indication.

OnLED      BIT     P1.3

; Appliaction RAM

APPFLAGS     DATA    21h
  TRQFLAG   BIT     APPFLAGS.0
                      ; Flag for monitoring I2C transmission success.
  SErrFLAG  BIT     APPFLAGS.1

FAILCNT:   DS      1

TOGCNT:    DS      1          ; Toggle counter.


;*****************************************************************************

;                                Program Start

;*****************************************************************************
           CSEG

; Reset and interrupt vectors.

           AJMP    Reset               ;Reset vector at address 0.


; A timer I timeout usually indicates a 'hung' bus.

           ORG     1Bh                 ; Timer I (I2C timeout) interrupt.
TimerI:    SETB    CLRTI
           AJMP    TIISR               ; Go to Interrupt Service Routine.


;******************************************************************************
;                    I2C Interrupt Service Routine
;******************************************************************************

; Notes on the interrupt mechanism:

; Other interrupts are enabled during this ISR upon return from XRETI.
; Limitations imposed on other ISR's:
; - Should not be long (close to 1000 clock cycles).  A long ISR will cause
;   the I2C bus to 'hang", and a TIMERI interrupt to occur.
; - Other interrupts either do not use the same mechanism for allowing
;   further interrupts, or if they do - disable TIMERI interrupt beforehand.

; The 751 hardware allows only one level of interrupts.  We simulate an
; additional level by software: by performing a RETI instruction (at location
; XRETI) the interrupt-in-progress flip-flop is cleared, and other interrupts
; are enabled.  The second level of interrupt is a must in our implementation,
; enabling timeout interrupts to occur during "stuck" wait loops in the I2C
; interrupt service routine.

           ORG     23h

I2CISR:    CLR     EI2         ; Disable I2C interrupt.
           ACALL   XRETI       ; Allow other interrupts to occur.
           PUSH    PSW
           PUSH    ACC
           MOV     A,R0
           PUSH    ACC
           MOV     A,R1
           PUSH    ACC
           MOV     A,R2
           PUSH    ACC

           MOV     StackSave, SP
           CLR     TIRUN
           SETB    TIRUN

           JB      STP,NoGo
           JNB     MASTER, GoSlave
           MOV     MSGSTAT,#MGO
           JB      STR,GoMaster
NoGo:      MOV     MSGSTAT,#NOTSTR
           AJMP    Dismiss             ; Not a valid Start.

XRETI:     RETI


;*****************************************************************************
;                    Main Transmit and Receive Routines
;*****************************************************************************

;  SLAVE CODE -
;  GET THE ADDRESS

GoSlave:   MOV     MSGSTAT,#SGO
AddrRcv:   ACALL   ClsRcv8
           JNB     DRDY, SMsgEnd       ; Must be some strange Start or Stop
                                       ; before the address byte was completed.
                                       ; Not a valid address.
STstRW:    MOV     C,ACC.0             ; Save R/W~ bit in carry.
           CLR     ACC.0               ; Clear that bit, leaving "raw" address
           JZ      GoIdle              ; If it is a General Address
                                       ; - ignore it.

                                       ; NOTE:
                                       ; One may insert here a different
                                       ; treatment for general calls, if
                                       ; these are relevant.

           JC      SlvTx               ; It's a Read - (requesting slave
                                       ; transmit).

; It is a Write (slave should receive the message).

;          Check if message is for us

SRcv2:     CJNE    A,MYADDR,GoIdle     ; If not my address - ignore the
                                       ; message.
           MOV     R1,#SRcvBuf         ; Set receive buffer address.
           MOV     R2,#RbufLen+1
           SJMP    SRcv3

SRcvSto:   MOV     @R1,A               ; Store the byte
           Inc     R1                  ; Step address.
SRcv3:     ACALL   AckRcv8
           JNB     DRDY,SRcvEnd        ; Exit loop -end reception.
           DJNZ    R2,SRcvSto          ; Go to store byte if buffer not full.

; Too many bytes received - do not acknowledge.
           MOV     MSGSTAT,#SRLNG      ; Notify main that (as slave) we
                                       ; have received too long a message.
           ACALL   SLnRCvdR            ; Handle new data - slave event routine.
           SJMP    GoIdle

; Received a byte, but not DRDY - check if a legitimate message end.

SRcvEnd:   CJNE    R0,#7,SRcvErr       ; If bit count not 7, it was not
                                       ; a Start or a Stop.

; Received a complete message

           MOV     MSGSTAT,#SRCVD
                                       ; Calculate number of bytes received
           MOV     A,R1
           CLR     C
           SUBB    A,#SRcvBuf          ; number of bytes in ACC
           ACALL   SRCvdR              ; Handle new data - slave event routine.
           SJMP    SMsgEnd

; It is a Read message, check if for us.

SlvTx:     NOP

STx2:      CJNE    A,MYADDR,GoIdle     ; Not for us.
           MOV     I2DAT,#0            ; Acknowledge the address.
           JNB     ATN,$               ; Wait for attention flag.
           JNB     DRDY,SMsgEnd        ; Exception - unexpected Start
                                       ; or Stop before the Ack got out.
           MOV     R1,#STxBuf          ; Start address of transmit buffer.
STxlp:     MOV     A,@R1               ; Get byte from buffer
           INC     R1
           ACALL   XmByte
           JNB     DRDY,SMsgEnd        ; Byte Tx not completed.
           JNB     RDAT,STxlp          ; Byte acknowledge, proceed trans.
           MOV     I2CON,#BCDR+BIDLE   ; Master Nak'ed for msg end.
           MOV     MSGSTAT,#STXED
           ACALL   STXedR              ; Slave transmitted event routine.
           AJMP    Dismiss

SRcvErr:   MOV     MSGSTAT,#SRERR      ; Flag bus/protocol error
           ACALL   SRErrR              ; Slave error event routine.
           SJMP    SMsgEnd
StxErr:    MOV     MSGSTAT,#SRERR      ; Flag bus/protocol error
           ACALL   SRErrR

SMsgEnd:   JB      MASTER,SMsgEnd2
           JB      STR,GoSlave         ; If it was a Start, be Slave
SMsgEnd2:  AJMP    Dismiss


; End of Slave message processing

GoIdle:    AJMP    Dismiss


GoMaster:

; Send address & R/W~ byte

           MOV     R1,#MasBuf          ; Master buffer address
           MOV     R2,MASTCNT          ; # of bytes, to send or rcv
           MOV     A,DESTADRW          ; Destination address (including
                                       ; R/W~ byte).
           JB      SUBADD,GoMas2       ; Branch if subaddress is needed.

           ACALL   XmAddr

           JNB     DRDY,GM2
           JNB     ARL,GM3
GM2:       AJMP    AdTxArl             ; Arbitration loss while transmitting
                                       ; the address.
GM3:       JB      RDAT,Noslave        ; No Ack for address transmission.
           JB      ACC.0, MRcv         ; Check R/W~ bit
           AJMP    MTx


; Handling subaddress case:

GoMas2:    NOP                         ; Subaddress needed. Address in ACC.
           CLR     ACC.0               ; Force a Write bit with address.
           ACALL   XmAddr
           JNB     DRDY,GM4
           JNB     ARL,GM5
GM4:       AJMP    AdTxArl             ; Arbitration loss while transmitting
                                       ; the address.

GM5:       JB      RDAT,Noslave        ; No Ack for address transmission.
           MOV     A,DESSUBAD
           ACALL   XmByte              ; Transmit subaddress.
           JNB     DRDY,SMsgEnd2       ; Arbitration loss (by Start or Stop)
           JB      ARL,SMsgEnd2        ; Arbitration loss occured.
           JB      RDAT,NoAck          ; Subadress transmission was not ack'ed.
           MOV     A,DESTADRW          ; Reload ACC with address.
           JNB     ACC.0, MTx          ; It's a Write, so proceed
                                       ; by sending the data.
; Read message, needs rp. Start and add. retransmit.

           MOV     I2CON,#BCDR+BXSTR   ; Send Repeated Start.
           JNB     ATN,$
           MOV     I2CON,#BCDR         ; Clear useless DRDY while preparing
                                       ; for Repeated Start.
           JNB     ATN,$               ; expecting an STR.
           JNB     ARL,GM6
           AJMP    MArlEnd             ; oops - lost arbitration.
GM6:       ACALL   XmAddr              ; Retransmit address, this time with the
                                       ; Read bit set.
           JNB     DRDY,GM7
           JNB     ARL,GM8
GM7:       AJMP    AdTxArl             ; Arbitration loss while transmitting
                                       ; the address.
GM8:       JB      RDAT,Noslave        ; No Ack - the slave disappeared.
           SJMP    MRcv                ; Proceed receiving slave's data.


; A Write message.  Master transmits the data.

MTx:       NOP

MTxLoop:   MOV     A,@R1               ; Get byte from buffer.
           INC     R1                  ; Step the address.
           ACALL   XmByte
           JNB     DRDY,SMsgEnd2       ; Arbitration loss (by Start or Stop)
           JB      ARL,SMsgEnd2        ; Arbitration loss.
           JB      RDAT,NoAck
           DJNZ    R2,MTxLoop          ; Loop if more bytes to send.

           MOV     MSGSTAT,#MTXED      ; Report completion of buffer
                                       ; transmission.
           SJMP    MTxStop
NoSlave:   MOV     MSGSTAT,#MTXNOSLV
           SJMP    MTxStop
NoAck:     MOV     MSGSTAT,#MTXNAK
           SJMP    MTxStop


; Master receive - a Read frame

MRcv:      ACALL   ClaRcv8             ; Receive a byte.
           SJMP    MRcv2
MRcvLoop:  ACALL   AckRcv8
MRcv2:     JNB     DRDY,MArl           ; Other's Start or Stop.
           MOV     @R1,A               ; Store received byte.
           INC     R1                  ; Advance address.
           DJNZ    R2,MRcvLoop

; Received the desired number of bytes - send Nack.

           MOV     I2DAT,#80h
           JNB     ATN,$
           JNB     DRDY,MArl
           MOV     MSGSTAT,#MRCVED
           SJMP    MTxStop             ; Go to send Stop or Repeated Start.


; Conclude this Master message:
; Send Stop, or a Repeated Start

MTxStop:   JNB     RPSTRT,MTxStop2     ; Check if Repeated Start needed
                                       ; Around if not RPSTRT.
           MOV     I2CON,#BCDR+BXSTR   ; Send Repeated Start.
           SJMP    MTxStop3
MTxStop2:  MOV     C,SETMRQ            ; Set new Master Request if demanded
           MOV     MASTRQ,C            ; by SETMRQ bit of MASCMD.

           MOV     I2CON,#BCDR+BXSTP   ; Request the HW to send a Stop.

MTxStop3:  JNB     ATN,$               ; Wait for Attention
           MOV     I2CON,#BCDR         ; Clear the useless DRDY, generated
                                       ; by SCL going high in preparation
                                       ; for thr Stop or Repeated Start.
           JNB     ATN,$               ; Wait for ARL, STP or STR.
           JB      ARL,MarlEnd         ; Lost arbitration trying to send
                                       ; Stop or a ReStart.

; Master is done with this message.  May proceed with new messages, if any,
; or exit.

           ACALL   MastNext            ; Master Event Routine.  May Prepare
                                       ; the pointers and data for the
                                       ; next Master message.

           JNB     MASTRQ,MMsgEnd      ; Go end service routine if MASTRQ
                                       ; does not indicate that the master
                                       ; should continue (was set according
                                       ; to SETMRQ bit, or by MastNext).

           JNB     STR,MMsgEnd         ; Return from the ISR, unless Start
                                       ; (avoid danger if we do not return:
                                       ; if there was a Stop, the watchdog
                                       ; is inactive until next Start).
           AJMP    GoMaster            ; Loop for another Master message

MMsgEnd:                               ;  End of Master messages,
           SJMP    Dismiss


; Terminate mastership due to an arbitration loss:

MArl:      JNB     STR,MArl2           ; If lost arbitration due to other
                                       ; Master's Start, go be a slave.
           AJMP    GoSlave

Marl2:     AJMP    Dismiss


; Switch from Master to Slave due to arbitration loss after completing
; transmission of a message.  The MASTRQ bit was cleared trying to write a
; Stop, and we need to set it again on order to retry transmission when the
; bus gets free again.

MArlEnd:   SETB    MASTRQ              ; Set Master Request - which will get
                                       ; into effect when we are done as a
                                       ; slave.
           AJMP    MArl

; Handling arbitration loss while transmitting an address

AdTxArl:   JB      STR,MArl            ; Non-synchronous Start or Stop.
           JB      STP,MArl

; Switch from Master to Slave due to arbitration loss while transmitting
; an address - complete receiving the address transmitted by the new Master.

           CJNE    R0,#0,AdTxArl2
                                       ; Arl on last bit of address
                                       ; (R0 is 0 on exit from XmAddr).
           DEC     A                   ; The lsb sent, in which arl occured
                                       ; must have been 1. By decrementing
                                       ; A we get the address that won.
           SJMP    AdAr3

AdTxArl2:  RR      A                   ; Realign partially Tx'ed ACC
           MOV     R1,A                ; and save itin R1
           MOV     A,R0                ; Pointer for lookup table
           MOV     DPTR,#MaskTable
           MOVC    A,@A+DPTR
           ANL     A,R1                ; Set address bits to be received,
                                       ; and the bit on which we lost
                                       ; arbitration to 0
                                       ; Now we are ready to receive the rest
                                       ; of the address.

           MOV     I2CON,#BCXA+BCARL   ; Clear flags and release the clock.

           ACALL   RBit3               ; Complete the address using reception
                                       ; subroutine.
           JB      DRDY,AdAr3          ; Around if received address OK
           AJMP    SMsgEnd             ; Unexpected Start or Stop - end
                                       ; as a slave.
AdAr3:     AJMP    STstRW              ; Proceed to check the address
                                       ; as a slave.

MaskTable:  DB      0ffh,7Eh,3Eh,1Eh,0Eh,06h,02h,00h, ; 0ffh is dummy

; End I2C Interrupt Service Routine:

Dismiss:   ACALL   I2CDONE

           MOV     I2CON,#BCARL+BCSTP+BCDR+BCXA+BIDLE
           CLR     TIRUN
           POP     ACC
           MOV     R2,A
           POP     ACC
           MOV     R1,A
           POP     ACC
           MOV     R0,A
           POP     ACC
           POP     PSW
           SETB    EI2

           RET                 ; Return from I2C interrupt Service Routine


;*****************************************************************************
;                Byte Transmit and Receive Subroutines
;*****************************************************************************

;  XmAddr: Transmit Address and R/W~
;  XmByte: Transmit a byte

XmAddr:    MOV     I2DAT,A             ; Send first bit, clears DRDY.
           MOV     I2CON,#BCARL+BCSTR+BCSTP
                                       ; Clear status, release SCL.
           MOV     R0,#8               ; Set R0 as bit counter
           SJMP    XmBit2
XmByte:    MOV     R0,#8
XmBit:     MOV     I2DAT,A             ; Send the first bit.
XmBit2:    RL      A                   ; Get next bit.
           JNB     ATN,$               ; Wait for bit sent.
           JNB     DRDY,XmBex          ; Should be data ready.
           DJNZ    R0,XmBit            ; Repeat until all bits sent.
           MOV     I2CON,#BCDR+BCXA    ; Switch to receive mode.
           JNB     ATN,$               ; Wait for acknowledge bit.
                                       ; flag cleared.
XmBex:     RET


; Byte receive routines.

; ClsRcv8  clears the status register (from Start condition)
;          and then receives a byte.
; AckRcv8  Sends an acknowledge, and then receives a new byte.
;          If a Start or Stop is encountered immediately after the
;          ack, AckRcv8 returns with 7 in R0.
; ClaRcv8  clears the transmit active state and releases clock
;          (from the acknowledge).

;          A contains the received byte upon return.
;          R0 is being used as a bit counter.

ClsRcv8:   MOV     I2CON,#BCARL+BCSTR+BCSTP+BCXA
                            ;Clear status register.
           JNB     ATN,$
           JNB     DRDY,RCVex
           SJMP    Rcv8

AckRcv8:   MOV     I2DAT,#0            ; Send Ack (low)
           JNB     ATN,$
           JNB     DRDY,RCVerr         ; Bus exception - exit.
ClaRcv8:   MOV     I2CON,#BCDR+BCXA    ; clear status, release clock
                                            ;from writing the Ack.  
           JNB     ATN,$

Rcv8:      MOV     R0,#7               ; Set bit counter for the first seven
                                       ; bits.
           CLR     A                   ; Init received byte to 0.
RBit:      ORL     A,I2DAT             ; Get bit, clear ATN.
RBit2:     RL      A                   ; Shift data.
           JNB     ATN,$               ; Wait for next bit.
           JNB     DRDY,RCVex          ; Exit if not a data bit (could be Start/
                                       ;  Stop, or bus/protocol error)
RBit3:     DJNZ    R0,RBit             ; Repeat until 7 bits are in.
           MOV     C,RDAT              ; Get last bit, don't clear ATN.
           RLC     A                   ; Form full data byte.
RCVex:     RET


RCVerr:    MOV     R0,#9               ; Return non legitimate bit count
           RET


;*****************************************************************************
;          Timer I Interrupt Service Routine
;          I2C us Timeout
;*****************************************************************************

; In addition to reporting the timeout in MSGSTAT, we update a failure
; counter, TITOCNT. This allows different types of timeout handling by the
; main program.

TIISR:     CLR     MASTRQ              ; "Manual" reset.
           MOV     I2CON,#BXSTP
           MOV     I2CON,#BCXA+BCDR+BCARL+BCSTR+BCSTP

TI1:       MOV     MSGSTAT,#TIMOUT     ; Status Flag for Main.
TI2:       MOV     A,#0FFh
           CJNE    A,TITOCNT,TI3       ; Increment TITOCNT, saturating
           SJMP    TI4                 ; at FFh.
TI3:       INC     TITOCNT

TI4:       ACALL   RECOVER

           SETB    CLRTI               ; Clear TI interrupt flag.
           ACALL   XRETI               ; Clear interrupt pending flag (in
                                       ; order to re-enable interrupts).
           MOV     SP,StackSave        ; Realign stack pointer, re-doing
                                       ; possible stack  changes during
                                       ; the I2C interrupt service routine.
                                       ; TimerI interrupts in other ISR's
                                       ; were not allowed !
           AJMP    Dismiss             ; Go back to the I2C service routine,
                                       ; in order to return to the (main)
                                       ; program interrupted.


;******************************************************************************
;       Bus recovery attempt subroutine
;******************************************************************************

RECOVER:   CLR     EA
           CLR     MASTRQ              ; "Manual" reset.
           MOV     I2CON,#BCXA+BIDLE+BCDR+BCARL+BCSTR+BCSTP
           CLR     SLAVEN     ; Non I2C TimerI mode
           SETB    TIRUN      ; Fire up TimerI.  When it overflows, it
                              ; will cause I2C interface hardware reset.
           MOV     R1,#0ffh
DLY5:      NOP
           NOP
           NOP
           DJNZ    R1,DLY5
           CLR     TIRUN
           SETB    CLRTI

           SETB    SCL        ; Issue clocks to help release other devices.
           SETB    SDA
           MOV     R1,#08h
RC7:       CLR     SCL
           DB      0,0,0,0,0
           SETB    SCL
           DB      0,0,0,0,0
           DJNZ    R1,RC7
           CLR     SCL
           DB      0,0
           CLR     SDA
           DB      0,0
           SETB    SCL
           DB      0,0,0,0,0
           SETB    SDA
           DB      0,0,0,0,0  ; Issue a Stop.

Rex:       MOV     I2CON,#BCXA+BCDR+BCARL+BCSTR+BCSTP ; clear flags
           SETB    EA
           RET


;*****************************************************************************

;                               Main Program

;*****************************************************************************

; Message ping pong game.  Each message is transmitted by
; a processor that is a master on the I2C bus, and it contains one byte
; of data.  A processor that receives this data byte as a slave increments
; the data by one and transmits it back as a master.  The data received is
; confirmed to be a one increment of the data formerly sent, unless
; it is a "reset" value, chosen to be 00h.
; The two participating processors have similar code, where the node
; address of the second processor is the destination address of this
; one, and vice versa.
; The first data byte each processor tries to send is 00h.  One of the
; processors will acquire the bus first, and the second processor that will
; receive this "resetting" 00h will not attempt tp confirm it against an
; expected value.  It will simply increment and transmit it.  Subsequent
; receptions will be confirmed against the expected value, until 0ffh data
; bytes are sent and the game is effectively reset by the 00h resulting from
; the next increment.
; A toggling output (TogLED) tells the outer world that the "ping pong"
; proceeds well.  If something unexpected happens we temporarily activate
; another output, ErrLED.
; The different tasks of the code are performed in a combination of main-
; line program and event routines called from the I2C interrupt service
; routine.


;  Initial set-ups:
;      Load CT1,CT0 bits of I2CFG register, according to the clock
;            crystal used.
;      Load RAM location MYADDR with the I2C address of this processor.
;    We load these values out of ROM table locations (R_CTVAL and R_MYADDR).
;    One may, instead, load with a MOV <immediate> command.

Reset:     MOV     SP,#07h             ;Set stack location.
           CLR     A
           MOV     DPTR,#R_CTVAL
           MOVC    A,@A+DPTR
           MOV     I2CFG,A             ; Load CT1,CT0 (I2C timing, crystal
                                       ; dependent).
           CLR     A
           MOV     DPTR,#R_MYADDR
           MOVC    A,@A+DPTR           ; Get this node's address from ROM table
           MOV     MYADDR,A            ; into MYADDR RAM location.

           CLR     OnLED

Reset2:    CLR     ErrLED              ; Flash LED.
           ACALL   LDELAY
           SETB    ErrLED
           CLR     SErrFLAG
           CLR     TRQFLAG
           MOV     FAILCNT,#50h
           SETB    TogLED
           MOV     TOGCNT,#050h        ; Initialize pin-toggling counter

; Enable slave operation.
; The Idle bit is set here for a restart situation - in normal
; operation this is redundant, as this bit is set upon power_up reset.
           MOV     I2CON,#BIDLE        ; Slave will idle till next Start.
           SETB    SLAVEN              ; Enable slave operation.

; Enable interrupts.
; This is necessary for both Slave and Master operations.
           SETB    ETI                 ; Enable timer I interrupts.
           SETB    EI2                 ; Enable I2C port interrupts.
           SETB    EA                  ; Enable global interrupts.

; Set up Master operation.

           MOV     MASCMD,#0h          ; "Regular" master transmissions.
           MOV     DPTR,#PongADDR
           CLR     A
           MOVC    A,@A+DPTR
           MOV     DESTADRW,A          ; The partner address.  The LSB is
                                       ; low, for a Write transaction.
           MOV     MASTCNT,#01h        ; Message length - a single byte.

PPSTART:   MOV     MasBuf,#00h

; "Ping" transmission:

PP2:       SETB    TRQFLAG
           SETB    MASTRQ
           MOV     R1,#0ffh
PP22:      JNB     TRQFLAG,PP3         ; Transmitted OK
           DJNZ    R1,PP22
MFAIL1:    DJNZ    FAILCNT,PP2
           ACALL   RECOVER
           SJMP    Reset2

; "Pong" reception:

PP3:       MOV     R0,#0ffh            ; Software timeout loop count.
PP31:      MOV     R1,#0ffh
PP32:      JB      TRQFLAG,PP2         ; Rcvd ok as slave, go transmit.
           JB      SErrFLAG,PP5
           DJNZ    R1,PP32
           DJNZ    R0,PP31
PPTO:      ACALL   RECOVER             ; Software timeout.
           AJMP    Reset2

PP5:       CLR     ErrLED              ; Receive error.
           ACALL   LDELAY
           SETB    ErrLED
           CLR     SErrFLAG
           AJMP    PPSTART

LDELAY:    MOV     R2,#030h
LDELAY1:   MOV     R1,#0ffh
           DJNZ    R1,$
           DJNZ    R2,LDELAY1
           RET


;*****************************************************************************
;  Slave and Master Event Routines.
;*****************************************************************************

;  Invoked upon completion of a message transaction.
;  This is the part of the application program actually dealing
;  with the data communicated on the I2C bus, by respoonding to
;  new data received and/or preparing the next transaction.

; Slave Event Routines

; These routines are invoked by the I2C interrupt service routine when a
; message transaction as a slave has been completed.  Our "application"
; reacts to a message received as a slave with the routine SRCvdR.
; The calls that indicate erronous reception are treated the same way as
; erronous data reception in the "ping pong" game.

;  SRcvdR
;          Invoked when a new message has been received as a Slave.

SRcvdR:    NOP
           MOV     A,SRcvBuf
           JNZ     SR2
           MOV     MasBuf,#01h          ; It was ping-pong reset value
           SJMP    SR3

SR2:       INC     MasBuf              ; The expected data.
           CJNE    A,MasBuf,ErrSR
           INC     MasBuf              ; Data for next transmission - the data
                                       ; received incremented by 1.

;          A successful two way data exchange.  Let the outside world know by
;          toggling an output pin driving a LED.  We actually toggle only
;          when a number of such exchanges is completed, in order to
;          slow down the changes for a good visual indication.

           DJNZ    TOGCNT,SR3
           CPL     TogLED              ; Toggle output
           MOV     TOGCNT,#050h

SR3:       CLR     SErrFLAG
           SETB    TRQFLAG             ; Request main to transmit
           RET


ErrSR:     SETB    SErrFLAG
           RET

;   SLnRcvdR
;          Invoked when a message received as a Slave is too long
;          for the receive buffer.

;   STXedR
;          Invoked when a Slave completed transmission of its buffer.
;          We do not expect to get here, since we do not plan to have
;          in our system a master that will request data from this node.

;   SRErrR
;          Slave error event subroutine.
;          In most applications it will not be used.


SLnRcvdR:
STXedR:
SRErrR:    JMP     ErrSR


;  MastNext - Master Event Routine.

;          Invoked when a Master transaction is completed, or terminated
;          "willingly" due to lack of acknowledge by a slave.

MastNext:  MOV     A,MSGSTAT
           CJNE    A,#MTXED,MN1
           MOV     FAILCNT,#50h
           CLR     TRQFLAG
           RET

MN1:       RET


; I2CDONE
;          Called upon completion of the I2C interrupt service routine.
;          In this example it monitors exceptions, and invokes the bus
;          recovery routine when too many occured.

I2CDONE:   MOV     A,MSGSTAT
           CJNE    A,#NOTSTR,I2CD1
           DJNZ    FAILCNT,I2CD1
           MOV     FAILCNT,#01h        ; Too many "illegal" i2c interrupts
           CLR     EI2                 ; - shut off.
I2CD1:     RET


;*****************************************************************************
;          I2C Communications Table:
;*****************************************************************************

; We used table driven values for clarity.  one may use immediates to load
; these values and save several lines of code.

; Contents is used in the begining of the main program to load
; RAM location MYADDR and the I2CFG register.
; The node address, in R_MYADDR, is application specific, and unique for
; each device in the I2C network.
; R_CTVAL depends on the crystal clock frequency.

R_MYADDR:   DB     4Eh         ; This node's address

R_CTVAL:    DB     02h         ; CT1, CT0 bit values


;*****************************************************************************
;          Application Code Definitions
;*****************************************************************************

PongADDR:  DB      4Ah        ; The address of the "partner" in
                              ; the ping-pong game.


           END

	